package polyglot.ext.ibex.lr;

import java.util.*;
import polyglot.main.Report;
import polyglot.util.InternalCompilerError;

/**
 * An abstract NFA.
 */
abstract class AbstractNFA implements NFA
{
    Map edges;
    List terminals;

    AbstractNFA(List terminals) {
        edges = new HashMap();
        this.terminals = terminals;
    }

    public void dump() {
        if (Report.should_report(TOPICS, 3)) {
            Report.report(1, "Dumping " + this);

            System.out.println();
            System.out.println("digraph G {");

            for (Iterator i = nodes(); i.hasNext(); ) {
                NFAState n = (NFAState) i.next();
                System.out.println("\"" + n + "\"");
                for (Iterator j = succs(n); j.hasNext(); ) {
                    NFAEdge e = (NFAEdge) j.next();
                    System.out.println("/* " + e + " */");
                    System.out.print("  \"" + e.source() + "\"");
                    System.out.print("->\"" + e.target() + "\"");
                    if (e.label() != null) {
                        System.out.println(" [ label=\"" + e.label() + "\" ]");
                    }
                    else {
                        System.out.println();
                    }
                }
            }

            System.out.println("}");
            System.out.println();
        }
    }

    public List alphabet() { return terminals; }

    public abstract boolean accepting(NFAState q0, NFAState q);
    public abstract BitSet accepting(NFAState q0);

    abstract NFAEdge edge(NFAState q1, Label a, NFAState q2);

    public abstract NFAState nodeAt(int i);
 
    public Iterator succs(Node n) {
        Collection out = (Collection) edges.get(n);
        if (out == null) {
            return Collections.EMPTY_LIST.iterator();
        }
        return out.iterator();
    }

    public Iterator preds(Node n) {
        throw new InternalCompilerError("unimplemented");
    }

    NFAEdge transition(NFAState q1, Label a, NFAState q2) {
        Collection c = (Collection) edges.get(q1);

        if (c == null) {
            c = new HashSet();
            edges.put(q1, c);
        }

        NFAEdge e = edge(q1, a, q2);
        c.add(e);

        return e;
    }

    NFAEdge epsilon(NFAState q1, NFAState q2) {
        return transition(q1, null, q2);
    }

    static Collection TOPICS = Arrays.asList(
                new String[] { "lr", "ibex", "nfa" });
}
