package polyglot.ext.ibex.lr;

import java.util.*;
import polyglot.main.Report;
import polyglot.util.InternalCompilerError;

class WrapperDFA implements NFA {
    NFA dfa;
    NFAState start;

    WrapperDFA(NFA dfa, NFAState start) {
        this.dfa = dfa;
        this.start = start;
        if (start == null) throw new InternalCompilerError("bad start");
    }

    public NFAState start() { return start; }
    public boolean accepting(NFAState q0, NFAState q) {
        return q0 == start && dfa.accepting(q0, q);
    }
    public BitSet accepting(NFAState q0) {
        if (q0 == start) return new BitSet();
        return dfa.accepting(q0);
    }
    public NFAState nodeAt(int i) { return dfa.nodeAt(i); }
    public List alphabet() { return dfa.alphabet(); }
    public int size() { return dfa.size(); }
    public Iterator succs(Node node) { return dfa.succs(node); }
    public Iterator preds(Node node) { return dfa.preds(node); }
    public Iterator nodes() { return dfa.nodes(); }

    public void dump() {
        System.out.println("start = " + start);
        dfa.dump();
    }

    public String toString() {
        return dfa.toString() + " start=" + start;
    }

    public boolean equals(Object o) {
        if (o instanceof WrapperDFA) {
            WrapperDFA m = (WrapperDFA) o;
            return start.equals(m.start) && dfa.equals(m.dfa);
        }
        return false;
    }
}
