package polyglot.ext.polyj.ast;

import polyglot.ext.jl.ast.*;

import polyglot.ext.polyj.types.*;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;

import java.util.*;

public class AmbInstOrAccess_c extends Node_c implements AmbInstOrAccess
{
    /**
     * "A" in "A[B]"
     */
    protected Receiver prefix;
    
    /**
     * "B" in "A[B]"
     */
    protected Receiver param;
    
    /**
     * The type of the node.
     */
    protected Type type;

    public AmbInstOrAccess_c(Position pos, Receiver prefix, Receiver param) {
        super(pos);
        this.prefix = prefix;
        this.param = param;
    }

    public Receiver prefix() {
        return this.prefix;
    }

    public AmbInstOrAccess prefix(Receiver prefix) {
        AmbInstOrAccess_c n = (AmbInstOrAccess_c) copy();
        n.prefix = prefix;
        return n;
    }

    public Receiver param() {
        return this.param;
    }

    public AmbInstOrAccess param(Receiver param) {
        AmbInstOrAccess_c n = (AmbInstOrAccess_c) copy();
        n.param = param;
        return n;
    }

    public Type type() {
        return this.type;
    }

    protected AmbInstOrAccess_c reconstruct(Receiver prefix, Receiver param) {
        if (prefix != this.prefix || param != this.param) {
            AmbInstOrAccess_c n = (AmbInstOrAccess_c) copy();
            n.prefix = prefix;
            n.param = param;
            return n;
        }
        return this;
    }

    public Node visitChildren(NodeVisitor v) {
        Receiver prefix = (Receiver) visitChild(this.prefix, v);
        Receiver param  = (Receiver) visitChild(this.param, v);
        return reconstruct(prefix, param);
    }

    public String toString() {
        return prefix + "[" + param + "]{amb}";
    }

    public Node disambiguate(AmbiguityRemover ar) throws SemanticException {
        PolyJTypeSystem ts = (PolyJTypeSystem) ar.typeSystem();
        PolyJNodeFactory nf = (PolyJNodeFactory) ar.nodeFactory();

        // Now that "prefix" has been disambiguated, we can tell whether
        // this is meant to be a type or an access.
        
        if (prefix instanceof TypeNode) {
            // "param" must be a parameter.
            TypeNode tn = (TypeNode) prefix;
            Type t = tn.type();
            PolyJClassType pt = null;

            if (! (t instanceof PolyJClassType)) {
                throw new SemanticException(tn.type() + " is not a parameterized type.", position());
            } else {
                pt = (PolyJClassType) t;
            }

            ParamNode disambParam = (ParamNode) nf.Param(param.position(), param);
            
            List l = new LinkedList();
            l.add(disambParam.parameter());

            Type instType = ts.instantiate(position(), pt, l);
            return nf.CanonicalTypeNode(position(), instType);
            
        } else if (prefix instanceof Expr) {
            // "param" must be an expression.
            if (param instanceof Expr) {
                return nf.ArrayAccess(position(), (Expr) prefix, (Expr) param);
            } else {
                throw new SemanticException("Illegal array access.", position());
            }
            
        }

        throw new SemanticException("Unable to disambiguate type or expression.", position());
    }
    
    public void prettyPrint(CodeWriter w, PrettyPrinter tr) {
        print(prefix, w, tr);
        w.write("[");
        print(param, w, tr);
        w.write("]");
    }

    public void translate(CodeWriter w, Translator tr) {
        throw new InternalCompilerError("Cannot translate " + this);
    }
}
