package polyglot.ext.polyj.ast;

import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.util.*;
import polyglot.ext.polyj.visit.*;

import polyglot.ext.jl.ast.*;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import java.util.*;

/**
 * A <code>PolyJCast</code> is an immutable representation of a casting
 * operation done in any PolyJ program.  It consists of an <code>Expr</code> being 
 * cast and a <code>Type</code> being cast to.
 **/
public class PolyJCast_c extends Cast_c implements PolyJCast {
    
    public PolyJCast_c(Position pos, TypeNode castType, Expr expr) {
        super(pos, castType, expr);
    }

    /** 
     * This method overrides the translate method of <code>Cast</code>
     * since casting can be done on parameter types in PolyJ, and this is translated
     * differently.
     **/
    public void translate(CodeWriter w, Translator tr) {
        Context c = tr.context();
        PolyJTypeSystem ts = (PolyJTypeSystem) tr.typeSystem();
        Type myType = castType.type();

        if (myType.isClass() && ts.needsWhereObject(myType.toClass())) {
            // If the type we are casting to has a where object
            // then we must cast using the method in the class.
            ClassType classType = myType.toClass();
            String name = PolyJTranslator.fullNameForClass(classType);
            String nameDollar = PolyJTranslator.fullNameDollar(classType);

            w.write("((" + name + ")" );
            printSubExpr(expr, w, tr);
            w.write( ").$cast$$" + nameDollar + "( ");
            w.begin(0);
            for (Iterator args = ts.allTypeArgs(classType).iterator();
                args.hasNext(); )
            {
                Type arg = (Type) args.next();
                w.write("\"" + arg + "\"");
                if (args.hasNext()) {
                    w.write(",");
                }
                w.allowBreak(0);
            }
            w.write(")");
            w.end();
        } else {
            w.begin(0);
            w.write("(");
            castType.del().translate(w, tr);
            w.write(")");
            w.allowBreak(2, "");
            
            printSubExpr(expr, w, tr);
            w.end();
        }
    }
    
    public void dump(CodeWriter w) {
        w.write( "POLYJ CAST ");
    }
}
