package polyglot.ext.polyj.util;

import polyglot.types.*;
import polyglot.util.*;
import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.util.*;
import polyglot.ext.polyj.visit.*;
import java.util.*;

public class WhereBinding_c implements WhereBinding {
    String name;
    boolean isInherited;
    protected ProcedureInstance pi;
    
    public WhereBinding_c(ParamType param, ProcedureInstance constraint) {
	// An optimization is possible here -- we can use
	// the where binding from the superclass adapter, if
	// it exists.  To do so, uncomment the following code,
	// then finish the methods at the end of this class.
	ParamType origParam = param; //findOrigParam(param, constraint);
	if (origParam == null) {
	    throw new InternalCompilerError("Unable to construct where binding"
	        + " for " + constraint);
	}
	ClassType origClass = origParam.constrainedClass();
	PolyJTypeSystem ts = (PolyJTypeSystem) origClass.typeSystem();
	
	String methodName = "";
	if (constraint instanceof MethodInstance) {
  	    MethodInstance mi = (MethodInstance) constraint;
	    methodName = mi.name();
	}
	
	pi = constraint;
	name = PolyJTranslator.adapterMethodName("where", origParam, 
	    origClass, ts, methodName);
	isInherited = !param.equals(origParam);
    }

    public String toString() {
        return name + (isInherited ? " (inherited)" : "");
    }
    
    public String name() {
	return name;
    }

    public boolean isInherited() {
	return isInherited;
    }
    
    public ProcedureInstance procedureInstance() {
	return pi;
    }
    
    public MethodInstance methodInstance() {
        if (isMethod()) {
            return (MethodInstance) pi;
        } else {
            return null;
        }
    }

    public ConstructorInstance constructorInstance() {
        if (isConstructor()) {
            return (ConstructorInstance) pi;
        } else {
            return null;
        }
    }

    public boolean isMethod() {
        return pi instanceof MethodInstance;
    }

    public boolean isConstructor() {
        return pi instanceof ConstructorInstance;
    }

    /**
     * Return the type parameter that was originally constrained by
     * <code>constraint</code>, where <code>param</code> is some
     * parameter also constrained by <code>constraint</code>;
     * or <code>null</code> if no such parameter can be found.
     * This basically tries to find the highest supertype that
     * would contain the where method for <code>constraint</code>.
     * For example, given these classes:
     * <code>
     * class A[T] where T { void foo(); }
     * class B[U] extends A[U] where U { void foo(); }
     * class C[V] extends B[V] where V { void foo(); }
     * </code>
     * and given <code>param = C#V</code> 
     * and <code>constraint = C#V.foo</code>
     * then this would return <code>A#T</code>.
     */ 
    protected ParamType findOrigParam(ParamType param, 
        ProcedureInstance constraint) 
    {
	//System.out.println("findOrigParam(" + param + ", " 
	//	+ constraint + ")");
	if (param == null) {
	    // Base case: no matching parameter could be found in a superclass.
	    //System.out.println("findOrigParam returns null for base case");
	    return null;
	}
	
	ClassType currClass = param.constrainedClass();
	ClassType superClass = currClass.superType().toClass();
	ParamType matchingParam = matchParameter(superClass, param);
	
	// Recurse
	ParamType origParam = findOrigParam(matchingParam, constraint);
	if (origParam != null) {
	    // The original param was found in some super type of currClass.
	    //System.out.println("findOrigParam returns " + origParam);
	    return origParam;
	}

	List procs;
	if (constraint instanceof MethodInstance) {
	    procs = param.methods();
	} else {
	    procs = param.constructors();
	}
	
	// *** If there's a bug, it'll be with the next line.
	if (containsSameProc(procs, constraint)) {
	    // No superclasses have the constrained parameter, and the
	    // current class does, so param is the originally 
	    // constrained parameter.
	    //System.out.println("findOrigParam returns " + param);
	    return param;
	}

	// Neither curClass nor its superclasses contain the constrained
	// parameter, so return failure.  This happens when we've
	// walked too high in the type hierarchy (i.e., up to Object,
	// or a class without a matching parameter).
	//System.out.println("findOrigParam returns null");
	return null;
    }
    
    /**
     * Return the parameter of <code>c</code> that matches 
     * <code>param</code>; or null if no such parameter can be found.
     * For example, given:
     * <code>
     * class A[X,Y,Z] extends B[int, X] {...}
     * </code>
     * <code>matchParameter(A#X)</code> will return <code>B#X</code>.
     */
    protected ParamType matchParameter(ClassType c, ParamType param) 
    {
	PolyJTypeSystem ts = (PolyJTypeSystem) c.typeSystem();
        Iterator argsIter = ts.allTypeArgs(c).iterator();
        Iterator paramsIter = ts.allTypeParams(c).iterator();
	while (argsIter.hasNext()) {
            Object a = argsIter.next();
            Object p = paramsIter.next();
            if (a.equals(param)) {
		return (ParamType) p;
	    }
        }
        return null;
    }
    
    protected boolean containsSameProc(List procs, ProcedureInstance pi) {
	/*
	for (Iterator i = procs.iterator(); i.hasNext(); ) {
	    ProcedureInstance p = (ProcedureInstance) i.next();
	    if (p.isSameProcedure(pi)) {
	        return true;
	    }
	}
	*/
	return false;
    }
}

